#FILE: blast_hit_flanking_sequence.pl
#AUTH: Paul Stothard (paul.stothard@gmail.com)
#DATE: July 2, 2008
#VERS: 1.0

use warnings;
use strict;
use Data::Dumper;
use File::Temp;
use Getopt::Long;
use Bio::Index::Fasta;
use Bio::Seq;

#initialize options
my %options = (
    blast_file  => undef,
    hit_file    => undef,
    query_file  => undef,
    outfile     => undef,
    downstream  => 100,
    upstream    => 100,
    n           => undef,
    hit_index   => undef,
    query_index => undef
);

#get command line options
GetOptions(
    'b=s' => \$options{blast_file},
    'h=s' => \$options{hit_file},
    'q=s' => \$options{query_file},
    'o=s' => \$options{outfile},
    'd=i' => \$options{downstream},
    'u=i' => \$options{upstream},
    'n=i' => \$options{n}
);

$options{upstream}   = get_positive_integer( $options{upstream} );
$options{downstream} = get_positive_integer( $options{downstream} );
$options{n}          = get_positive_integer( $options{n} );

#check for options
if (   !( defined( $options{blast_file} ) )
    or !( defined( $options{hit_file} ) )
    or !( defined( $options{query_file} ) )
    or !( defined( $options{outfile} ) )
    or !( defined( $options{downstream} ) )
    or !( defined( $options{upstream} ) ) )
{
    die( print_usage() . "\n" );
}

#build index file for sequences in the BLAST database
my $tmp_1      = new File::Temp();
my $filename_1 = $tmp_1->filename;
print "Building index file $filename_1 for $options{hit_file}\n";
$options{hit_index} = Bio::Index::Fasta->new( $filename_1, 'WRITE' );
$options{hit_index}->id_parser( \&id_parser );
$options{hit_index}->make_index( $options{hit_file} );
close($tmp_1) or die ("Cannot close file : $!");

#build index file for sequences in the query file
my $tmp_2      = new File::Temp();
my $filename_2 = $tmp_2->filename;
print "Building index file $filename_2 for $options{query_file}\n";
$options{query_index} = Bio::Index::Fasta->new( $filename_2, 'WRITE' );
$options{query_index}->id_parser( \&id_parser );
$options{query_index}->make_index( $options{query_file} );
close($tmp_2) or die ("Cannot close file : $!");

#get the blast results as an array of hashes
my @blast_results = @{ parse_blast_results( $options{blast_file} ) };

#print Dumper(@blast_results);

open( my $OUTFILE, '>', $options{outfile} )
    or die("Cannot open file '$options{outfile}': $!");

foreach (@blast_results) {
    my $hsp          = $_;
    my $new_title    = undef;
    my $new_sequence = undef;

    print "Processing query $hsp->{query_id}\n";

    #if no hsp was obtained, write out the query sequence
    if ( !defined( $hsp->{match_id} ) ) {
        $new_title    = $hsp->{query_id} . '_???';
        $new_sequence = uc(
            get_sequence_object_from_file_using_index( $options{query_index},
                $hsp->{query_id} )->seq()
        );
    }
    else {
        my $query_sequence_obj = get_sequence_object_from_file_using_index(
            $options{query_index}, $hsp->{query_id} );
        my $hit_sequence_obj
            = get_sequence_object_from_file_using_index( $options{hit_index},
            $hsp->{match_id} );

        my $query_portion_of_new_sequence
            = $query_sequence_obj->subseq( $hsp->{q_start}, $hsp->{q_end} );

        my $left_portion_of_new_sequence  = undef;
        my $right_portion_of_new_sequence = undef;

        my $left_amount  = $options{upstream};
        my $right_amount = $options{downstream};

        if ( $hsp->{hit_strand} == -1 ) {
            ( $left_amount, $right_amount ) = ( $right_amount, $left_amount );
        }

        my $left_end   = $hsp->{s_start} - 1;
        my $left_start = $left_end - $left_amount + 1 - $hsp->{q_start} + 1;
        if ( $left_end < 1 ) {
            $left_end = 1;
        }
        if ( $left_start < 1 ) {
            $left_start = 1;
        }

        if ( $left_end < $left_start ) {
            $left_portion_of_new_sequence = "";
        }
        else {
            $left_portion_of_new_sequence
                = $hit_sequence_obj->subseq( $left_start, $left_end );
        }

        my $right_start = $hsp->{s_end} + 1;
        my $right_end
            = $right_start 
            + $right_amount - 1
            + ( $query_sequence_obj->length - $hsp->{q_end} );
        if ( $right_end > $hit_sequence_obj->length ) {
            $right_end = $hit_sequence_obj->length;
        }
        if ( $right_start > $hit_sequence_obj->length ) {
            $right_start = $hit_sequence_obj->length;
        }

        if ( $right_end < $right_start ) {
            $right_portion_of_new_sequence = "";
        }
        else {
            $right_portion_of_new_sequence
                = $hit_sequence_obj->subseq( $right_start, $right_end );
        }

        #adjust flanking sequences according to $options{n} value
        if ( defined( $options{n} ) ) {
            my $excessive_n = $options{n} + 1;

            if ( $left_portion_of_new_sequence ne "" ) {
                my @left_split = split( /N{$excessive_n,}/i,
                    $left_portion_of_new_sequence );
                my $modified_left_portion_of_new_sequence = $left_split[-1];
                $left_start
                    = $left_start
                    + (
                          length($left_portion_of_new_sequence)
                        - length($modified_left_portion_of_new_sequence) );
                $left_portion_of_new_sequence
                    = $modified_left_portion_of_new_sequence;
            }

            if ( $right_portion_of_new_sequence ne "" ) {
                my @right_split = split( /N{$excessive_n,}/i,
                    $right_portion_of_new_sequence );
                my $modified_right_portion_of_new_sequence = $right_split[0];
                $right_end = $right_start
                    + length($modified_right_portion_of_new_sequence) - 1;
                $right_portion_of_new_sequence
                    = $modified_right_portion_of_new_sequence;
            }
        }

        if ( $hsp->{hit_strand} == 1 ) {
            $new_sequence
                = lc($left_portion_of_new_sequence)
                . uc($query_portion_of_new_sequence)
                . lc($right_portion_of_new_sequence);
        }

        #reverse complement if necessary
        if ( $hsp->{hit_strand} == -1 ) {
            $new_sequence
                = lc( reverse_complement($right_portion_of_new_sequence) )
                . uc($query_portion_of_new_sequence)
                . lc( reverse_complement($left_portion_of_new_sequence) );

            #$new_sequence = reverse_complement($new_sequence);
        }

        my $length = length($new_sequence);

        $new_title
            = "$hsp->{query_id};source=$hsp->{match_id};start=$left_start;end=$right_end;strand=$hsp->{hit_strand};length=$length;";

    }

    print {$OUTFILE} ">$new_title\n$new_sequence\n";
}

close($OUTFILE) or die("Cannot close file : $!");

sub reverse_complement {
    my $sequence = shift;
    $sequence
        =~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
    $sequence = reverse($sequence);
    return $sequence;
}

sub get_sequence_object_from_file_using_index {
    my $index = shift;
    my $title = lc(shift);

    if ( $title =~ m/gi\|(\d+)/ ) {
        $title = $1;
    }

    #returns a Bio::Seq object
    my $seqobj = $index->fetch($title);
    if ( !( defined($seqobj) ) ) {
        die("Could not obtain sequence for entry '$title'");
    }

    return $seqobj;
}

sub parse_blast_results {
    my $file = shift;

#the file can contain comments starting with'#'
#the file must have a line beginning with a 'query_id' indicating the column names:

    my $header_count = 0;
    my @column_names = ();
    my $columns_read = 0;

    open( my $INFILE, $file ) or die("Cannot open file '$file': $!");

    #obtain column names
    while ( my $line = <$INFILE> ) {
        $line =~ s/\cM|\n//g;
        $header_count++;
        if ( $line =~ m/^\#/ ) {
            next;
        }
        if ( $line =~ m/^query_id/ ) {
            $columns_read = 1;
            @column_names = @{ _split($line) };
            last;
        }
    }

    #read the remaining entries
    my @entries    = ();
    my $line_count = 0;
    while ( my $line = <$INFILE> ) {
        $line =~ s/\cM|\n//g;
        $line_count++;

        if ( $line =~ m/\S/ ) {
            my %entry  = ();
            my @values = @{ _split($line) };

            if ( scalar(@values) != scalar(@column_names) ) {

                #most likely not hit was obtained, i.e something like:
                #53_[T/C]_BTA-22554 no acceptable hits returned
                if ( scalar(@values) == 2 ) {
                    $entry{query_id} = $values[0];
                    push( @entries, \%entry );
                }

                next;
            }

            for ( my $i = 0; $i < scalar(@column_names); $i++ ) {
                $entry{ $column_names[$i] } = $values[$i];
            }

     #depending on the blast search, the s_start may be larger than the s_end,
     #to indicate that the query aligns to the reverse strand of the hit.
            if ( $entry{s_start} > $entry{s_end} ) {
                ( $entry{s_start}, $entry{s_end} )
                    = ( $entry{s_end}, $entry{s_start} );
                $entry{hit_strand} = -1;
            }
            else {
                $entry{hit_strand} = 1;
            }

            push( @entries, \%entry );
        }
    }
    close($INFILE) or die("Cannot close file : $!");
    return \@entries;
}

sub _split {
    my $line   = shift;
    my @values = ();
    if ( $line =~ m/\t/ ) {
        @values = split( /\t/, $line );
    }
    else {
        @values = split( /\,/, $line );
    }
    foreach (@values) {
        $_ = clean_value($_);
    }
    return \@values;
}

sub clean_value {
    my $value = shift;
    if ( !defined($value) ) {
        return ".";
    }
    if ( $value =~ m/^\s*$/ ) {
        return ".";
    }
    $value =~ s/^\s+//g;
    $value =~ s/\s+$//g;
    $value =~ s/\"|\'//g;
    return $value;
}

sub get_positive_integer {
    my $value = shift;
    my $int   = undef;
    if ( ( defined($value) ) && ( $value =~ m/(\d+)/ ) ) {
        $int = $1;
    }
    return $int;
}

sub id_parser {
    if ( $_[0] =~ /^>gi\|(\d+)/ ) {
        return lc($1);
    }
    elsif ( $_[0] =~ /^>([^\n]+)/ ) {
        return lc($1);
    }
    else {
        return;
    }
}

sub print_usage {
    print <<BLOCK;
USAGE: perl blast_hit_flanking_sequence.pl [-arguments]
 -b [FILE]     : the BLAST results file to process (Required).
 -h [FILE]     : FASTA file containing all database sequences
                 (Required).
 -q [FILE]     : FASTA file containing all query sequences (Required).
 -o [FILE]     : output file (Required).
 -d [INTEGER]  : the number of bases to add downstream of the
                 query. (Optional. Default is 100).
 -u [INTEGER]  : the number of bases to add upstream of the
                 query. (Optional. Default is 100).
 -n [INTEGER]  : maximum number of contiguous N characters
                 allowed. (Optional. Default is no maximum).

 perl blast_hit_flanking_sequence.pl -b blast_results.txt \
-h database.fasta -q queries.fasta -o out.fasta
BLOCK
}
